#include <math.h>
#include "lis.h"
#include "lis_solver.h"

int lis_gpbicg(LIS_VECTOR b, LIS_VECTOR x, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_VECTOR r, rtld, rhat, p, ptld, phat;
	LIS_VECTOR t, ttld, that, t0, t0hat;
	LIS_VECTOR y, w, u, z;
	LIS_SCALAR alpha, beta, rho, rho_old;
	LIS_SCALAR qsi, eta;
	LIS_SCALAR tmp, tmpdot[5];
	LIS_SCALAR bnrm2, nrm2;
	LIS_SCALAR tol, rhotol;
	int iter,maxiter,n,np,output;
	double times,ptimes;

	n       = A->n;
	np      = A->np + options[LIS_OPTIONS_PADDING]; 
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	tol     = params[LIS_PARAMS_RESID];

	rtld    = work[0];
	r       = work[1];
	rhat    = work[2];
	p       = work[3];
	ptld    = work[4];
	phat    = work[5];
	t       = work[6];
	ttld    = work[7];
	that    = work[8];
	t0      = work[9];
	t0hat   = work[10];
	y       = work[11];
	w       = work[12];
	u       = work[13];
	z       = work[14];

	alpha   = (LIS_SCALAR)1.0;
	qsi     = (LIS_SCALAR)1.0;
	rho_old = (LIS_SCALAR)1.0;

	ptimes  = 0.0;


	// Initial Residual
	if( lis_get_initial_residual(A,NULL,x,b,r,NULL,tol,&bnrm2,status) )
	{
		return LIS_SUCCESS;
	}
	lis_vector_copy(r,rtld);

	lis_vector_set_all(0,ttld);
	lis_vector_set_all(0,ptld);
	lis_vector_set_all(0,p);
	lis_vector_set_all(0,u);
	lis_vector_set_all(0,t);
	lis_vector_set_all(0,t0);
	
	for( iter=1; iter<=maxiter; iter++ )
	{
		/* rho = <rtld,r> */
		lis_vector_dot(rtld,r,&rho);

		/* test breakdown */
		if( rho==0.0 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_BREAKDOWN;
		}

		/* beta = (rho / rho_old) * (alpha / qsi) */
		beta = (rho / rho_old) * (alpha / qsi);

		/* w = ttld + beta*ptld */
		lis_vector_axpyz(beta,ptld,ttld,w);

		/* rhat = M^-1 * r */
		times = lis_wtime();
		lis_psolve(M, r, rhat);
		ptimes += lis_wtime()-times;

		/* p = rhat + beta*(p - u) */
		lis_vector_axpy(-1,u,p);
		lis_vector_xpay(rhat,beta,p);
		
		/* ptld = A * p */
		lis_matvec(A,p,ptld);

		/* tmpdot[0] = <rtld,ptld> */
		lis_vector_dot(rtld,ptld,&tmpdot[0]);
		/* test breakdown */
		/* */
		
		/* alpha = rho / tmpdot[0] */
		alpha = rho / tmpdot[0];

		/* y = t - r + alpha*(-w + ptld) */
		lis_vector_axpyz(-1,w,ptld,y);
		lis_vector_xpay(t,alpha,y);
		lis_vector_axpy(-1,r,y);

		/* t = r - alpha*ptld */
		lis_vector_axpyz(-alpha,ptld,r,t);

		/* that  = M^-1 * t */
		/* phat  = M^-1 * ptld */
		/* t0hat = M^-1 * t0 */
		times = lis_wtime();
		lis_psolve(M, t, that);
		lis_psolve(M, ptld, phat);
		lis_psolve(M, t0, t0hat);
		ptimes += lis_wtime()-times;

		/* ttld = A * that */
		lis_matvec(A,that,ttld);

		/* tmpdot[0] = <y,y>       */
		/* tmpdot[1] = <ttld,t>    */
		/* tmpdot[2] = <y,t>       */
		/* tmpdot[3] = <ttld,y>    */
		/* tmpdot[4] = <ttld,ttld> */
		lis_vector_dot(y,y,&tmpdot[0]);
		lis_vector_dot(ttld,t,&tmpdot[1]);
		lis_vector_dot(y,t,&tmpdot[2]);
		lis_vector_dot(ttld,y,&tmpdot[3]);
		lis_vector_dot(ttld,ttld,&tmpdot[4]);
		if(iter==1)
		{
			qsi = tmpdot[1] / tmpdot[4];
			eta = 0.0;
		}
		else
		{
			tmp = tmpdot[4]*tmpdot[0] - tmpdot[3]*tmpdot[3];
			qsi = (tmpdot[0]*tmpdot[1] - tmpdot[2]*tmpdot[3]) / tmp;
			eta = (tmpdot[4]*tmpdot[2] - tmpdot[3]*tmpdot[1]) / tmp;
		}

		/* u = qsi*phat + eta*(t0hat - rhat + beta*u) */
		lis_vector_xpay(t0hat,beta,u);
		lis_vector_axpy(-1,rhat,u);
		lis_vector_scale(eta,u);
		lis_vector_axpy(qsi,phat,u);

		/* z = qsi*rhat + eta*z - alpha*u */
		lis_vector_scale(eta,z);
		lis_vector_axpy(qsi,rhat,z);
		lis_vector_axpy(-alpha,u,z);

		/* x = x + alpha*p + z */
		lis_vector_axpy(alpha,p,x);
		lis_vector_axpy(1,z,x);
		
		/* r = t - eta*y - qsi*ttld */
		lis_vector_axpyz(-eta,y,t,r);
		lis_vector_axpy(-qsi,ttld,r);
		
		/* convergence check */
		lis_vector_nrm2(r,&nrm2);
		nrm2 = nrm2 * bnrm2;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}

		if( tol >= nrm2 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			status[LIS_STATUS_PTIMES] = ptimes;
			return LIS_SUCCESS;
		}

		lis_vector_copy(t,t0);
		rho_old = rho;
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	return LIS_MAXITER;
}
