#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#else
#ifdef WIN32
	#include <windows.h>
	#include <mmsystem.h>
	#pragma comment( lib, "winmm.lib" )
#else
#ifdef LINUX
	#include <sys/timeb.h>
#else
	#include <sys/time.h>
#endif
#endif
#endif
#include <time.h>
#include "lis.h"

void SETERR(char *fmt, ...)
{
	va_list vvlist;
	va_start( vvlist, fmt );
//	printf("%s: ",__func__);
	vprintf(fmt,vvlist);
	va_end(vvlist);
}

void CHKERR(int err)
{
	if(err)
	{
		lis_finalize();
		exit(err);
	}
}

#if defined(HAVE_MEMALIGN) || defined(HAVE_POSIX_MEMALIGN)

void *lis_malloc(size_t size)
{
#if defined(HAVE_MEMALIGN)
  return memalign(16, size);
#elif defined(HAVE_POSIX_MEMALIGN)
  return posix_memalign(16, size);
#endif
}

void lis_free(void *p)
{
  free(p);
  return;
}

#else
typedef struct _malloc_address {
  struct _malloc_address *next, *prev;
  void *address, *real_address;
  size_t size;
} malloc_address;

malloc_address malloc_address_top = 
  {&malloc_address_top, &malloc_address_top, NULL, NULL};


void *lis_malloc( size_t size )
{
  union {
    void *ptr;
    unsigned long l;
  } addr;
  malloc_address *ma;

	addr.ptr = malloc(size + 16);
    ma = malloc(sizeof(malloc_address));
    ma->next = &malloc_address_top;
	ma->prev = malloc_address_top.prev;
	ma->prev->next = ma;
	ma->next->prev = ma;
	ma->real_address = addr.ptr;
	ma->size = size;
	addr.l += 0xf;
	addr.l &= ~0xf;
	ma->address = addr.ptr;
 
  return addr.ptr;
}

void *lis_realloc( void *p, size_t size )
{
	malloc_address *ma;
	size_t old_size;
	void *s;

	ma = malloc_address_top.next;
	while (ma->address)
	{
	    if (ma->address == p) break;
		ma = ma->next;
	}
	if (ma->address)
	{
		old_size = ma->size;
		ma->next->prev = ma->prev;
		ma->prev->next = ma->next;
		s = lis_malloc(size);
		memcpy(s,p,old_size);
		free(ma->real_address);
		free(ma);
		return s;
	}
	s = realloc(p,size);
	return s;
}

void lis_free(void *p)
{
  malloc_address *ma;

  ma = malloc_address_top.next;
  while (ma->address) {
    if (ma->address == p) break;
	ma = ma->next;
  }
  if (ma->address) {
    ma->next->prev = ma->prev;
	ma->prev->next = ma->next;
    free(ma->real_address);
	free(ma);
	return;
  }
  free(p);
  return;
}
#endif

void lis_free2(int n, ...)
{
	va_list vvlist;
	int		i;
	char	*p;

	va_start( vvlist, n );
	for(i=0;i<n;i++)
	{
		p = va_arg(vvlist, char *);
		if( p )
		{
			lis_free(p);
		}
	}
	va_end(vvlist);
}

LIS_SCALAR lis_getbreak(void)
{
	LIS_SCALAR eps = 1.0;
	LIS_SCALAR x;
	LIS_SCALAR one = (LIS_SCALAR)1.0;

	do
	{
		eps = eps * 0.5;
		x   = one + eps;
	}while( x>one );

	return eps;
}

/*
double lis_cputime(void)
{
	clock_t ctime;
	
	ctime = clock();
	return (double)ctime / CLOCKS_PER_SEC;
}
*/

#ifdef _OPENMP
double lis_wtime(void)
{
	double  etime;
	
	etime = omp_get_wtime();

	return etime;
}
#else
#ifdef WIN32
static LARGE_INTEGER t = {0, 0};
double lis_wtime(void)
{

  LARGE_INTEGER l;
  if (t.QuadPart == 0)
    QueryPerformanceFrequency(&t);
  QueryPerformanceCounter(&l);
  return (double)l.QuadPart / (double)t.QuadPart;
  /*
  double  etime;
	
//	etime = (double)timeGetTime();
	etime = ((double)timeGetTime()) * 1.0e-3;

	return etime;
	*/
}
#else
#ifdef LINUX
double lis_wtime(void)
{
	double  etime;
	struct timeb tb;

	ftime(&tb);
	etime = tb.time + (double)tb.millitm*1.0e-3;
	return etime;
}
#else
double lis_wtime(void)
{
	double  etime;
	struct timeval tv;

	gettimeofday(&tv,NULL);
	etime = tv.tv_sec + (double)tv.tv_usec*1.0e-6;
	return etime;
}
#endif
#endif
#endif
